type FieldSchema = string | {
    description?: string;
    enum?: string[];
    type?: string;
};
type FieldsSchema<T = any> = {
    [K in keyof T]?: T[K] extends (infer U)[] ? ArrayFieldSchema<U> | string : T[K] extends object ? ObjectFieldSchema<T[K]> | string : FieldSchema;
};
type ArrayFieldSchema<T> = {
    description?: string;
    items?: FieldsSchema<T>;
} | string;
type ObjectFieldSchema<T> = {
    description?: string;
} & FieldsSchema<T>;

/**
 * The `JsonGPT` class offers methods for interacting with the OpenAI's functions model. Main feature is to fill empty values in objects.
 *
 * @template T The type of the target object for the AI model
 */
declare class JsonGPT {
    private _model;
    private _maxTokens;
    private _temperature;
    private _functionName;
    private _functionDescription;
    private _system;
    private _promptStartingObject;
    private _promptContent;
    private _target;
    private _prompt;
    private _content;
    private _topP;
    private _frequencyPenalty;
    private _presencePenalty;
    private _fieldsSchema;
    private _openAIApiKey;
    lastTarget: any;
    messages: any[];
    functions: any[];
    streaming: ReadableStream<any>;
    /**
     * The JsonGPT constructor accepts an optional target object.
     */
    constructor();
    /**
     * Set the prompt for the AI model
     * @param prompt - The prompt string
     * @returns - The `JsonGPT` instance for method chaining
     */
    prompt(prompt: string): this;
    /**
     * Set the name of the model.
     * It is suggested to use:
     * - "gpt-3.5-turbo-0613" (default)
     * - "gpt-4-0613"
     * - or any other model from https://platform.openai.com/docs/models/overview
     * @param name of the model
     */
    model(model: string): this;
    /**
     * Set the OpenAI key
     * @param openAIApiKey - Your OpenAI API Key
     */
    openAIApiKey(openAIApiKey: string): this;
    /**
     * Set the max number of tokens for the GPT model
     * The maximum number of to generate in the chat completion.
     * The total length of input tokens and generated tokens is limited by the model\'s context length.
     * @param maxTokens - Number of the max tokens
     */
    maxTokens(maxTokens: number): this;
    /**
     * Set the top_p parameter for the GPT model, a value between 0 and 1.
     * An alternative to sampling with temperature, called nucleus sampling,
     * where the model considers the results of the tokens with top_p probability mass.
     * So 0.1 means only the tokens comprising the top 10% probability mass are considered.
     * We generally recommend altering this or `temperature` but not both.
     * @param topP - The top_p parameter value
     */
    topP(topP: number): this;
    /**
     * Set the frequency penalty for the GPT model. Number between -2.0 and 2.0.
     * Positive values penalize new tokens based on their existing frequency in the text so far,
     * decreasing the model\'s likelihood to repeat the same line verbatim.
     * @param frequencyPenalty - The frequency penalty parameter value
     */
    frequencyPenalty(frequencyPenalty: number): this;
    /**
     * Set the presence penalty for the GPT model. Number between -2.0 and 2.0.
     * Positive values penalize new tokens based on whether they appear in the text so far,
     * increasing the model\'s likelihood to talk about new topics.
     * @param presencePenalty - The presence penalty parameter value
     */
    presencePenalty(presencePenalty: number): this;
    /**
     * Set the default startingObjec to add json object to the prompt
     * This is {target} JSON in promptContent
     */
    startingObject(obj: any): this;
    /**
     * Set the default prompt to add json object to the prompt
     *
     * Default: 'Starting function looks like this: ```{target}```.'
     * @param promptContent - The prompt to add content
     */
    promptStartingObject(promptStartingObject: string): this;
    /**
     * Set the default prompt to add content
     *
     * Default: 'The answer should be based on this content ```{content}```.'
     * @param promptContent - The prompt to add content
     */
    promptContent(promptContent: string): this;
    /**
     * Set the function name for the AI model
     *
     * Default: 'new_object_with_filled_data'
     * @param functionName - The function name
     */
    functionName(functionName: string): this;
    /**
     * Set the function description for the AI model
     *
     * Default: 'Function need all values to be filled to process.'
     * @param functionDescription - The function description
     */
    functionDescription(functionDescription: string): this;
    /**
     * Set the temperature for the GPT model. Value between 0 and 2.
     * What sampling temperature to use. Higher values like 0.8 will make the output more random,
     * while lower values like 0.2 will make it more focused and deterministic.
     * We generally recommend altering this or `top_p` but not both.
     *
     * @param temperature - The temperature parameter value
     */
    temperature(temperature: number): this;
    /**
     * Set the content of the AI's output. It is added to the prompt as an extra message with default prompt: 'The answer should be based on this content ```{content}```.'.
     * You can override it with promptContent function.
     *
     * @param content - Content of the AI's output
     */
    content(content: string): this;
    /**
     * Select Role or Character of the System (e.g. 'You are A software developer who can help you develop software programs and applications using programming languages and development tools. You need to call functions to complete your tasks.')
     *
     * Default: 'You are an AI assistant. You need to call functions to complete your tasks.'
     * @param system - Role or Character of the System
     */
    system(system: string): this;
    /**
     * Sets the field schema of the target object
     * @param fieldsSchema - Field schema of the target object
     */
    schema<T>(fieldsSchema: FieldsSchema<T>): this;
    /**
     * Asynchronously populate the AI input stream using the given parameters. To get an access to the stream, get a 'stream' property.
     * @param chunkHandler - An optional callback function to handle chunks of data
     * @returns - A Promise of type `JsonGPT`
     */
    stream(chunkHandler?: (chunk: any) => any): Promise<this>;
    private updateFunctions;
    private addPromptForStartingObject;
    private sendToOpenAIAPI;
    private prepareMessagesFirstMsg;
    private prepareMessagesForNext;
    private prepareMessageForFunctionCallAnswer;
}

type Callback = (chunk: string) => void;
declare function streamReader(stream: ReadableStream | Response, callback?: Callback): Promise<string>;

declare const _default: {
    JsonGPT: typeof JsonGPT;
};

export { JsonGPT, _default as default, streamReader };
