import { JsonAutoComplete } from "./utils/JsonAutoComplete";
import { isNode } from "./utils/envUtils";
import { FieldsSchema, getFunctionJsonSchema } from "./utils/objectsUtils";
import { IOpenAIBody, IOpenAIMsg, OpenAIStream } from "./utils/openAIStream";

/**
 * The `JsonGPT` class offers methods for interacting with the OpenAI's functions model. Main feature is to fill empty values in objects.
 *
 * @template T The type of the target object for the AI model
 */
class JsonGPT {
  // Private properties with default values
  private _model: string = "gpt-3.5-turbo-1106";
  private _maxTokens: number = 3097;
  private _temperature: number = 0;
  private _functionName: string = "new_object_with_filled_data";
  private _functionDescription: string = "Function need all values to be filled to process.";
  private _system: string = "You are an AI assistant. You need to call functions to complete your tasks.";
  private _promptStartingObject: string = "Starting function looks like this: ```{target}```.";
  private _promptContent: string = "The answer should be based on this content ```{content}```.";
  private _target: any;
  private _prompt: string;
  private _content: string;
  private _topP: number;
  private _frequencyPenalty: number;
  private _presencePenalty: number;
  private _fieldsSchema: FieldsSchema<any>;
  private _openAIApiKey;
  // Public properties, user can modify and get them - todo but should he modify, maybe better aproach to create getters?
  public lastTarget: any;
  public messages: any[];
  public functions: any[];
  public streaming: ReadableStream<any>;

  /**
   * The JsonGPT constructor accepts an optional target object.
   */
  constructor() {
    if (isNode()) this._openAIApiKey = process.env.OPENAI_API_KEY;
  }

  /**
   * Set the prompt for the AI model
   * @param prompt - The prompt string
   * @returns - The `JsonGPT` instance for method chaining
   */
  prompt(prompt: string): this {
    if (this._prompt) {
      this._prompt = prompt;
    } else {
      this._prompt = this.addPromptForStartingObject(this._target, prompt);
    }
    return this;
  }

  /**
   * Set the name of the model.
   * It is suggested to use:
   * - "gpt-3.5-turbo-0613" (default)
   * - "gpt-4-0613"
   * - or any other model from https://platform.openai.com/docs/models/overview
   * @param name of the model
   */
  model(model: string): this {
    this._model = model;
    return this;
  }

  /**
   * Set the OpenAI key
   * @param openAIApiKey - Your OpenAI API Key
   */
  openAIApiKey(openAIApiKey: string): this {
    this._openAIApiKey = openAIApiKey;
    return this;
  }

  /**
   * Set the max number of tokens for the GPT model
   * The maximum number of to generate in the chat completion.
   * The total length of input tokens and generated tokens is limited by the model\'s context length.
   * @param maxTokens - Number of the max tokens
   */
  maxTokens(maxTokens: number): this {
    this._maxTokens = maxTokens;
    return this;
  }

  /**
   * Set the top_p parameter for the GPT model, a value between 0 and 1.
   * An alternative to sampling with temperature, called nucleus sampling,
   * where the model considers the results of the tokens with top_p probability mass.
   * So 0.1 means only the tokens comprising the top 10% probability mass are considered.
   * We generally recommend altering this or `temperature` but not both.
   * @param topP - The top_p parameter value
   */
  topP(topP: number): this {
    this._topP = topP;
    return this;
  }

  /**
   * Set the frequency penalty for the GPT model. Number between -2.0 and 2.0.
   * Positive values penalize new tokens based on their existing frequency in the text so far,
   * decreasing the model\'s likelihood to repeat the same line verbatim.
   * @param frequencyPenalty - The frequency penalty parameter value
   */
  frequencyPenalty(frequencyPenalty: number): this {
    this._frequencyPenalty = frequencyPenalty;
    return this;
  }

  /**
   * Set the presence penalty for the GPT model. Number between -2.0 and 2.0.
   * Positive values penalize new tokens based on whether they appear in the text so far,
   * increasing the model\'s likelihood to talk about new topics.
   * @param presencePenalty - The presence penalty parameter value
   */
  presencePenalty(presencePenalty: number): this {
    this._presencePenalty = presencePenalty;
    return this;
  }

  /**
   * Set the default startingObjec to add json object to the prompt
   * This is {target} JSON in promptContent
   */
  startingObject(obj: any): this {
    this._target = obj;
    return this;
  }

  /**
   * Set the default prompt to add json object to the prompt
   *
   * Default: 'Starting function looks like this: ```{target}```.'
   * @param promptContent - The prompt to add content
   */
  promptStartingObject(promptStartingObject: string): this {
    this._promptStartingObject = promptStartingObject;
    return this;
  }

  /**
   * Set the default prompt to add content
   *
   * Default: 'The answer should be based on this content ```{content}```.'
   * @param promptContent - The prompt to add content
   */
  promptContent(promptContent: string): this {
    this._promptContent = promptContent;
    return this;
  }

  /**
   * Set the function name for the AI model
   *
   * Default: 'new_object_with_filled_data'
   * @param functionName - The function name
   */
  functionName(functionName: string): this {
    this._functionName = functionName;
    this.updateFunctions();
    return this;
  }

  /**
   * Set the function description for the AI model
   *
   * Default: 'Function need all values to be filled to process.'
   * @param functionDescription - The function description
   */
  functionDescription(functionDescription: string): this {
    this._functionDescription = functionDescription;
    this.updateFunctions();
    return this;
  }

  /**
   * Set the temperature for the GPT model. Value between 0 and 2.
   * What sampling temperature to use. Higher values like 0.8 will make the output more random,
   * while lower values like 0.2 will make it more focused and deterministic.
   * We generally recommend altering this or `top_p` but not both.
   *
   * @param temperature - The temperature parameter value
   */
  temperature(temperature: number): this {
    this._temperature = temperature;
    return this;
  }

  /**
   * Set the content of the AI's output. It is added to the prompt as an extra message with default prompt: 'The answer should be based on this content ```{content}```.'.
   * You can override it with promptContent function.
   *
   * @param content - Content of the AI's output
   */
  content(content: string): this {
    this._content = content;
    return this;
  }

  /**
   * Select Role or Character of the System (e.g. 'You are A software developer who can help you develop software programs and applications using programming languages and development tools. You need to call functions to complete your tasks.')
   *
   * Default: 'You are an AI assistant. You need to call functions to complete your tasks.'
   * @param system - Role or Character of the System
   */
  system(system: string): this {
    this._system = system;
    return this;
  }

  /**
   * Sets the field schema of the target object
   * @param fieldsSchema - Field schema of the target object
   */
  schema<T>(fieldsSchema: FieldsSchema<T>): this {
    this._fieldsSchema = fieldsSchema;
    this.updateFunctions();
    return this;
  }

  /**
   * Asynchronously populate the AI input stream using the given parameters. To get an access to the stream, get a 'stream' property.
   * @param chunkHandler - An optional callback function to handle chunks of data
   * @returns - A Promise of type `JsonGPT`
   */
  public async stream(chunkHandler?: (chunk: any) => any): Promise<this> {
    if (!this._fieldsSchema) throw Error("You need to provide schema");
    if (!this._target) throw Error("You need to provide target");
    if (!this._prompt) throw Error("You need to provide prompt");

    if (this.messages?.length > 0) {
      this.messages = [...this.messages, this.prepareMessagesForNext(this._prompt)];
    } else {
      this.messages = this.prepareMessagesFirstMsg(this._prompt, this._system, this._content);
    }

    // Create a Promise that resolves once the streamed responses are fully read and parsed
    const responseReady = new Promise<this>((resolve) => {
      //todo empty?
      //let handleChunk = async (chunk: string) => {};
      let handleChunk
      if (chunkHandler) {
        handleChunk = async (chunk: string) => {
          chunkHandler(chunk);
        };
      }

      this.sendToOpenAIAPI(this.messages, this.functions, handleChunk)
        .then((result: any) => {
          // Resolve the Promise with the final result
          resolve(result);
        })
        .catch((err) => {
          throw Error(err);
        });
    });

    return responseReady;
  }

  private updateFunctions() {
    if (this._target !== undefined && this._fieldsSchema !== undefined) {
      this.functions = [
        getFunctionJsonSchema(this._target, this._functionName, this._functionDescription, this._fieldsSchema),
      ];
    }
  }

  private addPromptForStartingObject(target: any, originalPrompt: string): string {
    return `${originalPrompt}. ${this._promptStartingObject.replace("{target}", JSON.stringify(target))}`;
  }

  private async sendToOpenAIAPI(
    messages: IOpenAIMsg[],
    functions: any[],
    chunkHandler?: (chunk: any) => void
  ): Promise<this> {
    if (!this._openAIApiKey) throw Error("You need to provide openAIApiKey");

    const payload: IOpenAIBody = {
      model: this._model,
      messages: messages,
      temperature: this._temperature,
      top_p: this._topP,
      frequency_penalty: this._frequencyPenalty,
      presence_penalty: this._presencePenalty,
      max_tokens: this._maxTokens,
      stream: true,
      function_call: { name: this._functionName },
      functions: functions,
      user: "JsonGPT.com",
      n: 1,
    };

    const jsAC = new JsonAutoComplete();

    //todo works but doesn't add messages
    this.streaming = await OpenAIStream(payload, this._openAIApiKey);
    if (!chunkHandler) return this;

    const reader = this.streaming.getReader();
    let resultFromOpenAI = "";

    const handleChunk = (chunk: string) => {
      resultFromOpenAI += chunk;
      jsAC.addChunk(chunk);
      let curr = jsAC.getCorrectJson();
      chunkHandler(curr);
    };

    const decoder = new TextDecoder();
    let done = false;

    while (!done) {
      const { value, done: doneReading } = await reader.read();
      done = doneReading;
      const chunkValue = decoder.decode(value);
        handleChunk(chunkValue);
      }

    this.messages = [...this.messages, this.prepareMessageForFunctionCallAnswer(resultFromOpenAI)];

    try {
      this.lastTarget = JSON.parse(resultFromOpenAI);
    } catch (ex) {
      new Error("Problem with parsing json object from OpenAI");
    }

    return this;
  }

  private prepareMessagesFirstMsg(prompt: string, system: string, content?: string): IOpenAIMsg[] {
    return [
      {
        role: "system",
        content: system,
      },
      {
        role: "user",
        content: `${prompt} ${content ? this._promptContent.replace("{content}", content) : ""} `,
      },
    ];
  }

  private prepareMessagesForNext(prompt: string): IOpenAIMsg {
    return {
      role: "user",
      content: prompt,
    };
  }

  private prepareMessageForFunctionCallAnswer(lastResult: string): IOpenAIMsg {
    return {
      role: "assistant",
      content: null,
      function_call: {
        name: this._functionName,
        arguments: lastResult,
      },
    };
  }
}

export default JsonGPT;
