type OpeningCharacter = "{" | "[" | '"';
type ClosingCharacter = "}" | "]" | '"';
const DICTIONARY: { [key: string]: ClosingCharacter } = {
  "{": "}",
  "[": "]",
  '"': '"',
};

export class JsonAutoComplete {
  private stack: ClosingCharacter[];
  private json: string;
  private correctJson: string;

  constructor() {
    this.stack = [];
    this.json = "";
    this.correctJson = "";
  }

  public addChunk(chunk: string): boolean {
    for (let i = 0; i < chunk.length; i++) {
      const char = chunk[i];
      const lastCharInStack = this.stack[this.stack.length - 1];

      // Handling quote characters requires special consideration because they can be used as attribute markers or as text
      if (char === '"' && (this.stack.length === 0 || lastCharInStack !== '"')) {
        // Adding opening quote to the stack
        this.stack.push(DICTIONARY[char]);
      } else if (char === '"' && lastCharInStack === '"') {
        // It's a closing quote
        this.stack.pop();
      } else if (DICTIONARY[char] && lastCharInStack !== '"') {
        // It's an opening bracket/brace outside of a string
        this.stack.push(DICTIONARY[char as OpeningCharacter]);
      } else if (char === lastCharInStack) {
        // It's a closing bracket/brace that matches the last in the stack
        this.stack.pop();
      }

      this.json += char;
    }

    // Attempt to parse the updated JSON to see if it is valid.
    try {
      let currentJson = this.getPossibleCorrectJson();
      JSON.parse(this.getPossibleCorrectJson()); // Use getCorrectJson to close any open brackets.
      this.correctJson = currentJson;
      return true;
    } catch (e) {
      // Parsing failed, indicating the chunk has led to invalid JSON. Revert the addition.
      return false;
    }
  }

  public getCorrectJson(): string {
    return this.correctJson;
  }

  private getPossibleCorrectJson() {
    // Copy the JSON so far and the stack
    let correctedJson = this.json;
    let stackCopy = [...this.stack];

    // Append necessary closing characters from the stack
    while (stackCopy.length > 0) {
      const closingChar = stackCopy.pop();
      correctedJson += closingChar;
    }

    return correctedJson;
  }
}
