import { describe, expect, it } from "vitest";
import { FieldsSchema, getFunctionJsonSchema } from "./objectsUtils";

describe("FieldDescriptor", () => {
  it("shortcut description", () => {
    const target = { age: 17 };
    const fieldsSchema: FieldsSchema<typeof target> = {
      age: "simple",
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    console.log(JSON.stringify(result, null, 2));
    expect(result.parameters.properties.age.description).toBe("simple");
  });

  it("all required", () => {
    const target = { age: 30, name: "Maciej", gender: "Male" };
    const fieldsSchema: FieldsSchema<typeof target> = {
      age: "my age",
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    //console.log(JSON.stringify(result, null, 2));
    expect(result.parameters.required).toEqual(["age", "name", "gender"]);
  });

  it("simple description", () => {
    const target = { age: 17 };
    const fieldsSchema: FieldsSchema<typeof target> = {
      age: {
        description: "simple",
      },
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    //console.log(JSON.stringify(result, null, 2));
    expect(result.parameters.properties.age.description).toBe("simple");
  });

  it("nested objects", () => {
    const target = {
      state: {
        name: "Dolnyslask",
        city: {
          name: "Wroclaw",
          street: {
            name: "Gliniana",
          },
        },
      },
    };
    const fieldsSchema: FieldsSchema<typeof target> = {
      state: {
        description: "this is state",
        name: "name of the state",
        city: {
          description: "this is city",
          name: "name of the city",
          street: "name of the street",
        },
      },
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    //console.log(JSON.stringify(result, null, 2));
    //todo dont want to run all of them, but each is success
    // expect(result.parameters.properties.state.description).toBe("this is state");
    // expect(result.parameters.properties.state.name.description).toBe("name of the state");
    // expect(result.parameters.properties.state.properties.city.description).toBe("this is city");
    // expect(result.parameters.properties.state.properties.city.name.description).toBe("name of the city");
    expect(result.parameters.properties.state.properties.city.properties.street.description).toBe("name of the street");
  });

  it("object with empty array", () => {
    //todo empty not working, but empty in object works
    const target = { elements: [] };
    const fieldsSchema: FieldsSchema<typeof target> = {
      elements: { description: "this is empty array" },
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    console.log(JSON.stringify(result, null, 2));
    expect(result.parameters.properties.elements.description).toBe("this is empty array");
  });

  it("object with array of objects", () => {
    //todo empty not working, but empty in object works
    const target = {
      elements: [
        {
          username: "asd",
          level: 2,
        },
      ],
    };

    const fieldsSchema: FieldsSchema<typeof target> = {
      elements: {
        description: "this is array",
        items: {
          username: "name of user",
          level: {
            description: "level of character",
          },
        },
      },
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    console.log(JSON.stringify(result, null, 2));
    //expect(result.parameters.properties.elements.items.properties.level.description).toBe("level of character");
  });

  it("object with array of strings", () => {
    //todo empty not working, but empty in object works
    const target = {
      tweets: ["My first tweet :) lets go!"],
    };

    const fieldsSchema: FieldsSchema<typeof target> = {
      tweets: "user tweets",
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    console.log(JSON.stringify(result, null, 2));
    expect(result.parameters.properties.tweets.description).toBe("user tweets");
  });

  it("object with enum parameter", () => {
    //todo empty not working, but empty in object works
    const target = {
      name: "Maciej",
      gender: "",
    };

    const fieldsSchema: FieldsSchema<typeof target> = {
      name: "name of person",
      gender: {
        description: "gender of person",
        enum: ["Female", "Male"],
      },
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    console.log(JSON.stringify(result, null, 2));
    expect(result.parameters.properties.gender.enum[0]).toBe("Female");
  });

  it("object with array that have enum parameter", () => {
    //todo empty not working, but empty in object works
    const target = {
      name: "Maciej",
      gender: "",
    };

    const fieldsSchema: FieldsSchema<typeof target> = {
      name: {
        description: "number",
        type:"number"
      },
      gender: {
        description: "gender of person",
        enum: ["Female", "Male"],
      },
    };
    const result = getFunctionJsonSchema(target, "function name", "description name", fieldsSchema);

    console.log(JSON.stringify(result, null, 2));
    expect(result.parameters.properties.gender.enum[0]).toBe("Female");
  });
});
