// Type for individual fields, can be a string or an object
type FieldSchema = string | { description?: string; enum?: string[]; type?: string };

// Type for objects and arrays
export type FieldsSchema<T = any> = {
  [K in keyof T]?: T[K] extends (infer U)[]
    ? ArrayFieldSchema<U> | string
    : T[K] extends object
    ? ObjectFieldSchema<T[K]> | string
    : FieldSchema;
};

// Type for array fields, can also be just a string
export type ArrayFieldSchema<T> = { description?: string; items?: FieldsSchema<T> } | string;

// Type for object fields
type ObjectFieldSchema<T> = { description?: string } & FieldsSchema<T>;

// Helper Type Guard function to check if descriptor is FieldDescriptor
function isFieldSchema(schema: string | FieldSchema): schema is FieldSchema {
  return typeof schema === "object";
}

// Handles error for unexpected data type
function handleUnexpectedDataType(data: any) {  
  throw Error(`Unexpected type of data: ${typeof data} with value: ${JSON.stringify(data)}`);
}

// Helper function to generate schema
function generateSchema(data: any, schema: FieldsSchema<any> | string = {}): any {
  //if null override with string
  if (data === null) {
    return { type: "string" };
  }

  if (Array.isArray(data)) {
    return {
      type: "array",
      items:
        data.length > 0 ? generateSchema(data[0], ((schema as any)[0] as FieldsSchema<any>) || {}) : { type: "string" },
    };
  }

  if (data instanceof Date) {
    return { type: "string", format: "date-time" };
  }

  if (typeof data === "object") {
    const finalJsonSchema: any = { type: "object", properties: {}, required: Object.keys(data) };

    for (let key in data) {
      if (schema && schema[key] !== undefined && isFieldSchema(schema[key])) {
        finalJsonSchema.properties[key] = {
          ...generateSchema(data[key], schema[key]),
          description: (schema[key] as any)?.description,
          enum: (schema[key] as any)?.enum,
        };
        //override type
        if ((schema[key] as any)?.type) {
          finalJsonSchema.properties[key].type = (schema[key] as any)?.type;
        }
      } else if (typeof schema === "object" && schema[key] && typeof schema[key] === "string") {
        finalJsonSchema.properties[key] = {
          ...generateSchema(data[key]),
          description: schema[key],
        };
      } else {
        finalJsonSchema.properties[key] = generateSchema(data[key]);
      }
    }
    return finalJsonSchema;
  }

  switch (typeof data) {
    case "number":
      return { type: "number" };
    case "boolean":
      return { type: "boolean" };
    case "string":
      return { type: "string" };
    default:
      handleUnexpectedDataType(data);
      return { type: "undefined" };
  }
}

export function getFunctionJsonSchema(target: any, name: string, description: string, fieldsSchema: FieldsSchema) {
  const schema = generateSchema(target, fieldsSchema);
  return {
    name,
    description,
    parameters: schema,
  };
}
