import { createParser, type ParsedEvent, type ReconnectInterval } from "eventsource-parser";

export interface IOpenAIBody {
  /**
   * ID of the model to use. table for details on which models work with the Chat API.
   */
  model: string;
  /**
   * A list of messages comprising the conversation so far.
   */
  messages: IOpenAIMsg[];
  /**
   * A list of functions the model may generate JSON inputs for.
   */
  functions?: any;
  /**
   * todo
   */
  function_call?: any;
  /**
   * What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic.  We generally recommend altering this or `top_p` but not both.
   */
  temperature?: number | null;
  /**
   * An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10% probability mass are considered.  We generally recommend altering this or `temperature` but not both.
   */
  top_p?: number | null;
  /**
   * How many chat completion choices to generate for each input message.
   */
  n?: number | null;
  /**
   * If set, partial message deltas will be sent, like in ChatGPT. Tokens will be sent as data-only as they become available, with the stream terminated by a `data: [DONE]` message.
   */
  stream?: boolean | null;
  /**
   * todo
   */
  stop?: Array<string> | string;
  /**
   * The maximum number of to generate in the chat completion.  The total length of input tokens and generated tokens is limited by the model\'s context length. 
   */
  max_tokens?: number;
  /**
   * Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model\'s likelihood to talk about new topics.
   */
  presence_penalty?: number | null;
  /**
   * Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model\'s likelihood to repeat the same line verbatim.
   */
  frequency_penalty?: number | null;
  /**
   * Modify the likelihood of specified tokens appearing in the completion.  Accepts a json object that maps tokens (specified by their token ID in the tokenizer) to an associated bias value from -100 to 100. Mathematically, the bias is added to the logits generated by the model prior to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the relevant token.
   */
  logit_bias?: object | null;
  /**
   * A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse.
   */
  user?: string;
}

export interface IOpenAIMsg {
  /**
   * The role of the messages author. One of `system`, `user`, `assistant`, or `function`.
   */
  role: "system" | "user" | "assistant" | "function";
  /**
   * The contents of the message. `content` is required for all messages except assistant messages with function calls.
   */
  content?: string | null;
  /**
   * The name of the author of this message. `name` is required if role is `function`, and it should be the name of the function whose response is in the `content`. May contain a-z, A-Z, 0-9, and underscores, with a maximum length of 64 characters.
   */
  name?: string;
  /**
   * Array of functions
   */
  function_call?: any;
}

export async function OpenAIStream(payload: IOpenAIBody, openAIApiKey: string) {
  const encoder = new TextEncoder();
  const decoder = new TextDecoder();

  const func_call = {
    name: null,
    arguments: "",
  };

  const res = await fetch("https://api.openai.com/v1/chat/completions", {
    headers: {
      "Content-Type": "application/json",
      Authorization: `Bearer ${openAIApiKey}`,
    },
    method: "POST",
    body: JSON.stringify(payload),
  });

  const stream = new ReadableStream({
    async start(controller) {
      function onParse(event: ParsedEvent | ReconnectInterval) {
        if (event.type === "event") {
          const data = event.data;

          if (data === "[DONE]") {
            controller.close();
            return;
          }

          try {
            const json = JSON.parse(data);

            if (json.choices && json.choices.length > 0) {
              const choice = json.choices[0];
              const delta = choice.delta;

              if (delta.function_call) {
                if (delta.function_call.name) {
                  func_call.name = delta.function_call.name;
                }
                if (delta.function_call.arguments) {
                  func_call.arguments += delta.function_call.arguments;
                }
              }

              if (choice.finish_reason === "function_call") {
                return;
              }

              //we only need arguments
              const text = delta?.function_call?.arguments || "";

              const queue = encoder.encode(text);
              controller.enqueue(queue);
            }
          } catch (e) {
            controller.error(e);
          }
        }
      }

      const parser = createParser(onParse);

      const reader = res?.body?.getReader();
      if (reader) {
        while (true) {
          const { done, value } = await reader.read();
          if (done) {
            break;
          }
          parser.feed(decoder.decode(value));
        }
      }

      // todo check the difference
      // for await (const chunk of res.body as any) {
      //   parser.feed(decoder.decode(chunk));
      // }
    },
  });

  return stream;
}
